import os
import subprocess

from jadi import component
from aj.plugins.services.api import ServiceManager, Service


INIT_D = '/etc/init.d'

@component(ServiceManager)
class SysVServiceManager(ServiceManager):
    """
    Manager for sysv init scripts.
    """

    id = 'sysv'
    name = 'System V'

    @classmethod
    def __verify__(cls):
        """
        Test if sysv init scripts are used

        :return: Existence of /etc/init.d
        :rtype: bool
        """

        return os.path.exists(INIT_D)

    def __init__(self, context):
        pass

    def list(self):
        """
        Generator of all scripts under /etc/init.d.

        :return: Service object
        :rtype: Service
        """

        for _id in os.listdir(INIT_D):
            path = os.path.join(INIT_D, _id)
            if _id.startswith('.'):
                continue
            if _id.startswith('rc'):
                continue
            if os.path.islink(path):
                continue
            if os.path.exists(f'/etc/init/{_id}.conf'):
                continue
            yield self.get_service(_id)

    def get_service(self, _id):
        """
        Get status for one specified init script.

        :param _id: Script name
        :type _id: string
        :return: Service object
        :rtype: Service
        """

        svc = Service(self)
        svc.id = svc.name = _id
        try:
            svc.running = self._run_action(_id, 'status')
            svc.state = 'running' if svc.running else 'stopped'
            svc.enabled = True
            svc.static = False
        except Exception as e:
            svc.running = False
        return svc

    def _run_action(self, _id, action):
        """
        Wrapper for basic scripts actions ( restart, start, stop, status ).

        :param _id: Script name
        :type _id: string
        :param action: Action ( restart, start, stop, status )
        :type action: string
        """

        return subprocess.call([os.path.join(INIT_D, _id), action], close_fds=True) == 0

    def start(self, _id):
        """
        Basically start a script.

        :param _id: Script name
        :type _id: string
        """

        self._run_action(_id, 'start')

    def stop(self, _id):
        """
        Basically stop a script.

        :param _id: Script name
        :type _id: string
        """

        self._run_action(_id, 'stop')

    def restart(self, _id):
        """
        Basically restart a script.

        :param _id: Script name
        :type _id: string
        """

        self._run_action(_id, 'restart')



from jadi import interface


class Service():
    """
    Basic class to store service informations.
    """

    def __init__(self, manager):
        self.id = None
        self.name = None
        self.manager = manager
        self.state = None
        self.running = None


class ServiceOperationError(Exception):
    """
    Exception class for services.
    """

    def __init__(self, inner):
        self.inner = inner

    def __unicode__(self):
        return f'[ServiceOperationError {self.inner}]'


@interface
class ServiceManager():
    """
    Abstract interface for all managers.
    """

    id = None
    name = None

    def list(self):
        raise NotImplementedError

    def get_service(self, _id):
        raise NotImplementedError

    def start(self, _id):
        raise NotImplementedError

    def stop(self, _id):
        raise NotImplementedError

    def restart(self, _id):
        raise NotImplementedError

    def kill(self, _id):
        raise NotImplementedError



"""
Module to manage the init services programs (sysv init and systemd).
"""

from jadi import component
from aj.api.http import get, post, HttpPlugin
from aj.auth import authorize
from aj.api.endpoint import endpoint, EndpointError
from aj.plugins.services.api import ServiceManager, ServiceOperationError


@component(HttpPlugin)
class Handler(HttpPlugin):
    def __init__(self, context):
        self.context = context
        self.managers = {x.id:x for x in ServiceManager.all(self.context)}

    def __service_to_json(self, svc):
        """
        Utility to convert a Service object into dict for angular.

        :param svc: Service object
        :type svc: Service
        :return: Services informations
        :rtype: dict
        """

        return {
            'id': svc.id,
            'name': svc.name,
            'state': svc.state,
            'running': svc.running,
            'managerId': svc.manager.id,
            'enabled': svc.enabled,
            'static': svc.static,
        }

    @get(r'/api/services/managers')
    @endpoint(api=True)
    def handle_api_managers(self, http_context):
        """
        List all available managers.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :return: List of managers, one manager per dict
        :rtype: list of dict
        """

        return [
            {
                'id': mgr.id,
                'name': mgr.name,
            } for mgr in self.managers.values()
        ]

    @get(r'/api/services/manager/(?P<manager_id>\w+)')
    @endpoint(api=True)
    def handle_api_list(self, http_context, manager_id=None):
        """
        Retrieve all services from one specified init system.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param manager_id: Manager id, e.g. systemd
        :type manager_id: string
        :return: List of services informations
        :rtype: list of dict
        """

        return [self.__service_to_json(svc) for svc in self.managers[manager_id].list()]

    @get(r'/api/services/manager/(?P<manager_id>\w+)/(?P<service_id>.+)')
    @endpoint(api=True)
    def handle_api_get(self, http_context, manager_id=None, service_id=None):
        """
        Retrieve the service informations for one specified service in one
        specified manager.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param manager_id: Manager id, e.g. systemd
        :type manager_id: string
        :param service_id: Service id, e.g. ssh
        :type service_id: string
        :return: Service informations
        :rtype: dict
        """

        return self.__service_to_json(self.managers[manager_id].get_service(service_id))

    @get(r'/api/services/status/(?P<manager_id>\w+)/(?P<service_id>.+)')
    @endpoint(api=True)
    def handle_api_get_status(self, http_context, manager_id=None, service_id=None):
        """
        Retrieve the service informations for one specified service in one
        specified manager.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param manager_id: Manager id, e.g. systemd
        :type manager_id: string
        :param service_id: Service id, e.g. ssh
        :type service_id: string
        :return: Service informations
        :rtype: dict
        """

        return self.managers[manager_id].get_status(service_id)

    @post(r'/api/services/do/(?P<operation>\w+)/(?P<manager_id>\w+)/(?P<service_id>.+)')
    @authorize('services:manage')
    @endpoint(api=True)
    def handle_api_operate(self, http_context, manager_id=None, operation=None, service_id=None):
        """
        Launch one command for a specified service.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param manager_id: Manager id, e.g. systemd
        :type manager_id: string
        :param operation: Operation type, e.g. start
        :type operation: string
        :param service_id: Service id, e.g. ssh
        :type service_id: string
        :return: Service informations
        :rtype: dict
        """

        if operation not in ['start', 'stop', 'restart', 'kill', 'disable', 'enable']:
            return
        try:
            getattr(self.managers[manager_id], operation)(service_id)
        except ServiceOperationError as e:
            raise EndpointError(e)
